// ConsoleApplication.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "stdlib.h"
#include "windows.h"
#include "mmsystem.h"
#include "dxutil.h"
#include "dsutil.h"

typedef enum {OK, FAIL} ErrorStatus;

#define AUDIO_BLOCK_SIZE 1000
#define WAVE_DISPLAY_DATA_BLOCK_LEN 40
#define MAX_SAMPLING_VALUE 8000

// the background thread state type
typedef enum {IDLE_STATE, ON_STATE, NORMAL_REC_STATE} TerminalStateType;

// the incoming command type
typedef enum {QUIT, 
              TURN_ON, 
			  TURN_OFF, 
			  START_NORMAL_REC, 
			  STOP_NORMAL_REC,
			  SET_ECHO,
			  SET_MUSIC_VOLUME,
			  SET_VOICE_VOLUME,
			  SET_FREQ,
              SET_DISTORTION} TerminalCmdType;

// the command type to the MCU
typedef enum {MCU_TURN_ON, 
              MCU_TURN_OFF, 
			  MCU_SET_ECHO,
			  MCU_SET_MUSIC_VOLUME,
			  MCU_SET_VOICE_VOLUME,
			  MCU_SET_FREQ,
              MCU_SET_DISTORTION} McuCmdType;

// the command packet format
typedef struct McuCmdPktTag {
	unsigned short int cmd;
	unsigned short int dataLen;
} McuCmdPkt;

// a single parameter command packet format
typedef struct SingleMcuCmdPktTag {
	McuCmdPkt basic;
	unsigned short param;
} SingleMcuCmdPkt;

// internal buffer to store the audio data
BYTE *sampleData;
// the CWaveFile class for wave file creation
CWaveFile waveFile;
// the background thread state
TerminalStateType terminalState;
// the COM port device handler
HANDLE hComPort;
// debugging file handler
FILE *hLogFile;
// current sampling rate
int samplingValue=8000;
int sampleIdx=0;
// the critical section
CRITICAL_SECTION critSection;
bool isQuit;


// function to create an empty wave file
ErrorStatus createWaveFile(LPTSTR pFilename, CWaveFile &waveFile) {
	WAVEFORMATEX waveFormat;

	waveFormat.wFormatTag=WAVE_FORMAT_PCM;
	waveFormat.nBlockAlign=1;
	waveFormat.nChannels=1;
	waveFormat.nSamplesPerSec=samplingValue;
	waveFormat.wBitsPerSample=8;
	waveFormat.nAvgBytesPerSec=samplingValue;
	waveFormat.cbSize=0L;

	HRESULT ret=waveFile.Open(pFilename, &waveFormat, WAVEFILE_WRITE);
	if (FAILED(ret)) {
		return FAIL;
	}

	return OK;
}

// set sampling rate value
void setSamplingValue(int value) {
	samplingValue=value;
}

// close the wave file 
ErrorStatus closeWaveFile(CWaveFile &waveFile) {
	HRESULT ret=waveFile.Close();
	if (FAILED(ret)) {
		return FAIL;
	}
	return OK;
}

// append audio data into the wave file
ErrorStatus appendData(BYTE *pBuf, DWORD nBytes, CWaveFile &waveFile)
{
	UINT bytesWrite;
	HRESULT ret=waveFile.Write(nBytes,pBuf,&bytesWrite);
	if (FAILED(ret)) {
		return FAIL;
	}

	return OK;
}

// print the serial port configuration settings
void printPortCfg(DCB &cfgPort)
{
	fprintf(hLogFile,"COM Port Configuration:\n");
	fprintf(hLogFile,"Baud Rate: %d\n",(int)cfgPort.BaudRate);
	fprintf(hLogFile,"Byte Size: %d\n",(int)cfgPort.ByteSize);
	fprintf(hLogFile,"DCB length: %d\n",(int)cfgPort.DCBlength);
	fprintf(hLogFile,"Eof Char: %d\n",(int)cfgPort.EofChar);
	fprintf(hLogFile,"Error Char: %d\n",(int)cfgPort.ErrorChar);
	fprintf(hLogFile,"Evt Char: %d\n",(int)cfgPort.EvtChar);
	fprintf(hLogFile,"fAbortOnError: %d\n",(int)cfgPort.fAbortOnError);
	fprintf(hLogFile,"fBinary: %d\n",(int)cfgPort.fBinary);
	fprintf(hLogFile,"fDsrSensitivity: %d\n",(int)cfgPort.fDsrSensitivity);
	fprintf(hLogFile,"fDtrControl: %d\n",(int)cfgPort.fDtrControl);
	fprintf(hLogFile,"fDummy2: %d\n",(int)cfgPort.fDummy2);
	fprintf(hLogFile,"fErrorChar: %d\n",(int)cfgPort.fErrorChar);
	fprintf(hLogFile,"fInX: %d\n",(int)cfgPort.fInX);
	fprintf(hLogFile,"fNull: %d\n",(int)cfgPort.fNull);
	fprintf(hLogFile,"fOutX: %d\n",(int)cfgPort.fOutX);
	fprintf(hLogFile,"fOutxCtsFlow: %d\n",(int)cfgPort.fOutxCtsFlow);
	fprintf(hLogFile,"fOutxDsrFlow: %d\n",(int)cfgPort.fOutxDsrFlow);
	fprintf(hLogFile,"fParity: %d\n",(int)cfgPort.fParity);
	fprintf(hLogFile,"fRtsControl: %d\n",(int)cfgPort.fRtsControl);
	fprintf(hLogFile,"fTXContinueOnXoff: %d\n",(int)cfgPort.fTXContinueOnXoff);
	fprintf(hLogFile,"Parity: %d\n",(int)cfgPort.Parity);
	fprintf(hLogFile,"StopBits: %d\n",(int)cfgPort.StopBits);
	fprintf(hLogFile,"wReserved: %d\n",(int)cfgPort.wReserved);
	fprintf(hLogFile,"wReserved1: %d\n",(int)cfgPort.wReserved1);
	fprintf(hLogFile,"XoffChar: %d\n",(int)cfgPort.XoffChar);
	fprintf(hLogFile,"XoffLim: %d\n",(int)cfgPort.XoffLim);
	fprintf(hLogFile,"XonChar: %d\n",(int)cfgPort.XonChar);
	fprintf(hLogFile,"XonLim: %d\n",(int)cfgPort.XonLim);
	fprintf(hLogFile,"\n");
	fflush(hLogFile);
}
// process the command coming from the Java application
ErrorStatus processCmd(int cmdIdx) {
	McuCmdPktTag mcuCmdPkt;
	SingleMcuCmdPkt singleMcuCmdPkt;
	DWORD nWrites;
	unsigned int param1;

	switch (cmdIdx) {
		case TURN_ON:
			// handle the "TURN_ON" command
			fprintf(hLogFile,"TURN_ON\n");
			fflush(hLogFile);
			mcuCmdPkt.cmd = MCU_TURN_ON;
			mcuCmdPkt.dataLen = 0;

			// send the corresponding command packet to the MCU
			if (WriteFile(hComPort, // handle to file to write to
						  &mcuCmdPkt,        // pointer to data to write to file
						  sizeof(mcuCmdPkt), // number of bytes to write
						  &nWrites,
						  NULL) == 0)      // pointer to number of bytes written
			{
				fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
				return FAIL;
			}
			sampleData=(BYTE *)malloc(MAX_SAMPLING_VALUE*sizeof(BYTE));
			sampleIdx = 0;
			terminalState=ON_STATE;

			break;
		case TURN_OFF:
			// handle the "TURN_OFF" command
			fprintf(hLogFile,"TURN_OFF\n");
			fflush(hLogFile);
			mcuCmdPkt.cmd = MCU_TURN_OFF;
			mcuCmdPkt.dataLen = 0;

			// send the corresponding command packet to the MCU
			if (WriteFile(hComPort, // handle to file to write to
						  &mcuCmdPkt,        // pointer to data to write to file
						  sizeof(mcuCmdPkt), // number of bytes to write
						  &nWrites,
						  NULL) == 0)      // pointer to number of bytes written
			{
				fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
				fflush(hLogFile);
				return FAIL;
			}

			free(sampleData);
			terminalState=IDLE_STATE;
			break;
		case START_NORMAL_REC:
			// handle the "START_NORMAL_REC" command
			fprintf(hLogFile,"START_NORMAL_REC\n");
			fflush(hLogFile);
			if (terminalState == NORMAL_REC_STATE) {
				fprintf(hLogFile,"The terminal has already started recording song!\n");
				fflush(hLogFile);
			} else {
				char fileName[256];
				gets(fileName);
				gets(fileName);
				fprintf(hLogFile,"Filename: %s\n",fileName);
				fflush(hLogFile);

				// create an empty wave file 
				if (createWaveFile(fileName, waveFile)) {
					fprintf(hLogFile,"Fail to create the sound file!\n");
					fflush(hLogFile);
					return FAIL;
				}
				
				terminalState=NORMAL_REC_STATE;

			}
			break;
		case STOP_NORMAL_REC:
			// handle the "STOP_NORMAL_REC" command
			fprintf(hLogFile,"STOP_NORMAL_REC\n");
			fflush(hLogFile);
			if (terminalState != NORMAL_REC_STATE) {
				fprintf(hLogFile,"The terminal has not started recording song!\n");
				fflush(hLogFile);
			} else {

				// close the wave file
				if (closeWaveFile(waveFile)) {
					fprintf(hLogFile,"Fail to close the sound file!\n");
					fflush(hLogFile);
					return FAIL;
				}

				terminalState=ON_STATE;
			}
			break;
		case SET_ECHO:
			// handle the "SET_ECHO" command
			fprintf(hLogFile,"SET_ECHO\n");
			fflush(hLogFile);
			scanf("%d",&param1);
			singleMcuCmdPkt.basic.cmd = MCU_SET_ECHO;
			singleMcuCmdPkt.basic.dataLen = 1;
			singleMcuCmdPkt.param = param1;
			fprintf(hLogFile,"SCALE: %d\n",param1);
			fflush(hLogFile);

			// send the corresponding command packet to the MCU
			if (WriteFile(hComPort, // handle to file to write to
						  &singleMcuCmdPkt,        // pointer to data to write to file
						  5, // number of bytes to write
						  &nWrites,
						  NULL) == 0)      // pointer to number of bytes written
			{
				fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
				fflush(hLogFile);
				return FAIL;
			}

			break;

		case SET_MUSIC_VOLUME:
			// handle the "SET_MUSIC_VOLUME" command
			fprintf(hLogFile,"SET_MUSIC_VOLUME\n");
			fflush(hLogFile);
			scanf("%d",&param1);
			singleMcuCmdPkt.basic.cmd = MCU_SET_MUSIC_VOLUME;
			singleMcuCmdPkt.basic.dataLen = 1;
			singleMcuCmdPkt.param = param1;
			// send the corresponding command packet to the MCU
			if (WriteFile(hComPort, // handle to file to write to
						  &singleMcuCmdPkt,        // pointer to data to write to file
						  5, // number of bytes to write
						  &nWrites,
						  NULL) == 0)      // pointer to number of bytes written
			{
				fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
				fflush(hLogFile);
				return FAIL;
			}

			break;

		case SET_VOICE_VOLUME:
			// handle the "SET_VOICE_VOLUME" command
			fprintf(hLogFile,"SET_VOICE_VOLUME\n");
			fflush(hLogFile);
			scanf("%d",&param1);
			singleMcuCmdPkt.basic.cmd = MCU_SET_VOICE_VOLUME;
			singleMcuCmdPkt.basic.dataLen = 1;
			singleMcuCmdPkt.param = param1;

			// send the corresponding command packet to the MCU
			if (WriteFile(hComPort, // handle to file to write to
						  &singleMcuCmdPkt,        // pointer to data to write to file
						  5, // number of bytes to write
						  &nWrites,
						  NULL) == 0)      // pointer to number of bytes written
			{
				fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
				fflush(hLogFile);
				return FAIL;
			}

			break;

		case SET_FREQ:
			// handle the "SET_FREQ" command
			if (terminalState == ON_STATE || terminalState == NORMAL_REC_STATE) {
				fprintf(hLogFile,"Cannot change the sampling rate when the system is in on state or rec state\n",GetLastError());
				fflush(hLogFile);
			} else {
				fprintf(hLogFile,"SET_FREQ\n");
				fflush(hLogFile);
				scanf("%d",&param1);
				singleMcuCmdPkt.basic.cmd = MCU_SET_FREQ;
				singleMcuCmdPkt.basic.dataLen = 1;
				singleMcuCmdPkt.param = param1;

				// send the corresponding command packet to the MCU
				if (WriteFile(hComPort, // handle to file to write to
							  &singleMcuCmdPkt,        // pointer to data to write to file
							  5, // number of bytes to write
							  &nWrites,
							  NULL) == 0)      // pointer to number of bytes written
				{
					fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
					fflush(hLogFile);
					return FAIL;
				}
				setSamplingValue(param1*1000);
			}
			break;
		case SET_DISTORTION:
			// handle the "SET_DISTORTION" command
			fprintf(hLogFile,"SET_DISTORTION\n");
			fflush(hLogFile);
			scanf("%d",&param1);
			singleMcuCmdPkt.basic.cmd = MCU_SET_DISTORTION;
			singleMcuCmdPkt.basic.dataLen = 1;
			singleMcuCmdPkt.param = param1;
			fprintf(hLogFile,"Type: %d\n",param1);
			fflush(hLogFile);

			// send the corresponding command packet to the MCU
			if (WriteFile(hComPort, // handle to file to write to
						  &singleMcuCmdPkt,        // pointer to data to write to file
						  5, // number of bytes to write
						  &nWrites,
						  NULL) == 0)      // pointer to number of bytes written
			{
				fprintf(hLogFile,"Writing of serial communication has problem. %d\n",GetLastError());
				fflush(hLogFile);
				return FAIL;
			}
			
			break;
	}
	return OK;
}

// function to handle the background task
ErrorStatus processBackgroundTask() {
	switch(terminalState) {
		case IDLE_STATE:
			break;
		case ON_STATE:
		case NORMAL_REC_STATE:
			{
				DWORD nReads;
				fprintf(hLogFile,"Wait for incoming data!\n");
				fflush(hLogFile);

				// read the real-time audio data from MCU via serial port
				if (ReadFile(hComPort, // handle to file to write to
							  sampleData+sampleIdx*AUDIO_BLOCK_SIZE,  // pointer to data to write to file
							  AUDIO_BLOCK_SIZE*sizeof(unsigned char), // number of bytes to write
							  &nReads,
							  NULL) == 0)      // pointer to number of bytes written
				{
					fprintf(hLogFile,"Reading of serial communication has problem. %d\n",GetLastError());
					fflush(hLogFile);
					return FAIL;
				}
				fprintf(hLogFile,"Received  incoming data!\n");
				fflush(hLogFile);
				/*
				// DEBUG
				for (int i=0;i<AUDIO_BLOCK_SIZE;i++) {
					*(sampleData+sampleIdx*AUDIO_BLOCK_SIZE+i)=(unsigned char)(((double)rand())/RAND_MAX*255);
				}*/

				if (terminalState ==  NORMAL_REC_STATE) {
					// append the audio data to the wave file if necessary
					if (appendData(sampleData+sampleIdx*AUDIO_BLOCK_SIZE, AUDIO_BLOCK_SIZE, waveFile)) {
						fprintf(hLogFile,"Fail to append data to the wave file!\n");
						fflush(hLogFile);
						return FAIL;
					}
				}

				sampleIdx++;

				// send the ddownsampled audio data to the PC
				if  (sampleIdx == samplingValue/AUDIO_BLOCK_SIZE) {
					unsigned char *currSample=sampleData;
					for (int i=0;i<WAVE_DISPLAY_DATA_BLOCK_LEN;i++) {
						
						unsigned int avg=0;
						unsigned int numSamplesForAvg=samplingValue/WAVE_DISPLAY_DATA_BLOCK_LEN;

						/*
						for (int j=0;j<numSamplesForAvg;j++) {
							avg+=(*currSample);
							currSample++;
						}
						
						avg/=numSamplesForAvg;
						*/
						avg=*currSample;
						currSample+=numSamplesForAvg;

						printf("%u\n",avg);
						fflush(stdout);
					}
					sampleIdx=0;
				}
				//fprintf(hLogFile,"sampleIdx: %d\n",sampleIdx);
				//fflush(hLogFile);

			}
			break;

	}
	return OK;
}

// the thread call back for the background task
DWORD WINAPI backgroundTaskCallBack(LPVOID lpParam) {
	while (!isQuit) {
		EnterCriticalSection(&critSection);

		if (processBackgroundTask()) {
			LeaveCriticalSection(&critSection);

			return -1;
		}
		LeaveCriticalSection(&critSection);
				Sleep(125);
	}
	return 0;
}

// the main thread
int _tmain(int argc, _TCHAR* argv[])
{
	if (argc != 2) {
		printf("Usage SerialPorts [com port #]\n");
		return -1;
	}
	// open the debugging file
	hLogFile = fopen("console_app_debug.txt","w");
	int comPort = atoi(argv[1]);

	fprintf(hLogFile,"Connecting port COM%d ....\n",comPort);
	//fprintf(hLogFile,"Number of seconds to be recorded %d!\n",numSecs);

	// open a serial port
	char portId[24];

	sprintf(portId,"COM%d",comPort);

	hComPort = CreateFile(portId,  // Specify port device: default "COM1"
                              GENERIC_READ | GENERIC_WRITE,       // Specify mode that open device.
							  0,                                  // the devide isn't shared.
							  NULL,                               // the object gets a default security.
							  OPEN_EXISTING,                      // Specify which action to take on file. 
							  0,                                  // default.
							  NULL); 


	DCB cfgPort;

	// Get current configuration of serial communication port.
    if (GetCommState(hComPort,&cfgPort) == 0)
    {
      fprintf(hLogFile,"Get configuration port has problem.");
      fflush(hLogFile);
      return -1;
    }

	// Assign user parameter.
    cfgPort.BaudRate = 115200; //CBR_57600;  // Specify buad rate of communicaiton.
    cfgPort.StopBits = ONESTOPBIT;  // Specify stopbit of communication.
    cfgPort.Parity = NOPARITY;      // Specify parity of communication.
    cfgPort.ByteSize = 8;  // Specify  byte of size of communication.
    cfgPort.fOutxCtsFlow = FALSE;
	cfgPort.fOutxDsrFlow = FALSE;
	cfgPort.fDtrControl  = FALSE;
    cfgPort.fRtsControl = FALSE;
	cfgPort.fOutX = FALSE;
	cfgPort.fInX = FALSE;

	printPortCfg(cfgPort);

	// set the serial port setting
	if (SetCommState(hComPort,&cfgPort) == 0)
    {
       fprintf(hLogFile,"Set configuration port has problem.");
       fflush(hLogFile);

       return -1;
    }


    // instance an object of COMMTIMEOUTS.
    COMMTIMEOUTS comTimeOut;                   
    // Specify time-out between charactor for receiving.
    comTimeOut.ReadIntervalTimeout = 0;//3;
    // Specify value that is multiplied 
    // by the requested number of bytes to be read. 
    comTimeOut.ReadTotalTimeoutMultiplier = 0;//3;
    // Specify value is added to the product of the 
    // ReadTotalTimeoutMultiplier member
    comTimeOut.ReadTotalTimeoutConstant = 0;//2;
    // Specify value that is multiplied 
    // by the requested number of bytes to be sent. 
    comTimeOut.WriteTotalTimeoutMultiplier = 0;//3;
    // Specify value is added to the product of the 
    // WriteTotalTimeoutMultiplier member
    comTimeOut.WriteTotalTimeoutConstant = 0;//2;
    // set the time-out parameter into device control.
    SetCommTimeouts(hComPort,&comTimeOut);
	TerminalCmdType cmdIdx;
	terminalState=IDLE_STATE;

	InitializeCriticalSection(&critSection);

	isQuit = false;

//	HANDLE hConsoleIn=GetStdHandle(STD_INPUT_HANDLE);

	HANDLE hThread;
	DWORD threadId;

	// create the thread to handle background task
	hThread=CreateThread(NULL,0,backgroundTaskCallBack,NULL,0,&threadId);

	if (hThread == NULL) {
		fprintf(hLogFile,"Fail to create a background thread!\n");
		fflush(hLogFile);
	
		return -1;
	}

	// loop for command from the Java application
	while (!isQuit) {

//		unsigned int ret =WaitForSingleObjectEx(hConsoleIn,0,false);
//		if (ret!=WAIT_TIMEOUT) {

		scanf("%d",&cmdIdx);
		fprintf(hLogFile,"Cmd Received: %d\n",cmdIdx);
		fflush(hLogFile);

		// acquire critical section to maintain mutual exclusion
		EnterCriticalSection(&critSection);
		if (cmdIdx==0) {
			isQuit=true;
		} else {
			if (processCmd(cmdIdx)) {
				LeaveCriticalSection(&critSection);
				return -1;
			}
		}
		// release critical section
		LeaveCriticalSection(&critSection);
//		}

	}

	// close the serial port
	if(CloseHandle(hComPort) == 0)    // Call this function to close port.
    {
      fprintf(hLogFile,"Port Closing isn't successed.");
	  fflush(hLogFile);
      return -1;
    }

	// close the debugging file
	fclose(hLogFile);

	return 0;
}

